#![allow(dead_code)]
use std::fs::File;
use std::io::{Read};
use std::process::Command;
use std::time::Duration;
use std::path::Path;

use clap::Parser;
use log::{info, warn, error};
use serde::{Deserialize, Serialize};
use tokio::time::sleep;

static PATH_FAN: &str = "/proc/acpi/ibm/fan";

#[derive(Parser)]
#[command(name = "thinkfand")]
#[command(about = "ThinkPad fan control daemon")]
struct Args {
    #[arg(short, long, default_value = "/etc/thinkfand.conf")]
    config: String,
    
    #[arg(short, long)]
    daemon: bool,
    
    #[arg(short, long)]
    level: Option<u8>,
    
    #[arg(short, long)]
    auto: bool,
    

}

#[derive(Debug, Clone)]
struct Input {
    name: String,
    temp: f64,
}

#[derive(Debug, Clone)]
struct Adapter {
    name: String,
    inputs: Vec<Input>,
}

#[derive(Debug, Serialize, Deserialize)]
struct FanConfig {
    sensors: Vec<String>,
    levels: Vec<FanLevel>,
}

#[derive(Debug, Serialize, Deserialize)]
struct FanLevel {
    temp: f64,
    level: u8,
}

struct ThinkFanDaemon {
    config: FanConfig,
    current_level: u8,
    auto_mode: bool,
}

impl ThinkFanDaemon {
    fn new(config_path: &str) -> anyhow::Result<Self> {
        let config = if Path::new(config_path).exists() {
            let content = std::fs::read_to_string(config_path)?;
            serde_json::from_str(&content)?
        } else {
            FanConfig {
                sensors: vec!["thinkpad-isa-0000".to_string()],
                levels: vec![
                    FanLevel { temp: 0.0, level: 0 },
                    FanLevel { temp: 50.0, level: 2 },
                    FanLevel { temp: 60.0, level: 4 },
                    FanLevel { temp: 70.0, level: 6 },
                    FanLevel { temp: 80.0, level: 7 },
                ],
            }
        };

        Ok(Self {
            config,
            current_level: 0,
            auto_mode: true,
        })
    }

    fn read_temperatures(&self) -> Vec<Adapter> {
        let output = match Command::new("sensors").arg("-j").output() {
            Ok(output) => output,
            Err(e) => {
                error!("Failed to run sensors command: {}", e);
                return vec![];
            }
        };

        let json_str = match std::str::from_utf8(&output.stdout) {
            Ok(s) => s,
            Err(e) => {
                error!("Invalid UTF-8 in sensors output: {}", e);
                return vec![];
            }
        };

        parse_adapters(json_str)
    }

    fn get_max_temp(&self, adapters: &[Adapter]) -> f64 {
        let mut max_temp: f64 = 0.0;
        
        for adapter in adapters {
            for input in &adapter.inputs {
                if self.config.sensors.iter().any(|s| adapter.name.contains(s)) {
                    max_temp = max_temp.max(input.temp);
                }
            }
        }
        
        max_temp
    }

    fn calculate_fan_level(&self, temp: f64) -> u8 {
        let mut level = 0;
        
        for fan_level in &self.config.levels {
            if temp >= fan_level.temp {
                level = fan_level.level;
            }
        }
        
        level
    }

    fn set_fan_level(&mut self, level: u8) -> anyhow::Result<()> {
        if self.current_level == level {
            return Ok(());
        }

        let command = format!("level {}", level);
        match std::fs::write(PATH_FAN, &command) {
            Ok(_) => {
                self.current_level = level;
                info!("Fan level set to {}", level);
                Ok(())
            }
            Err(e) => {
                let error_str = e.to_string();
                if error_str.contains("Invalid argument") {
                    warn!("Invalid argument error, reloading thinkpad_acpi module");
                    self.reload_thinkpad_acpi()?;
                    
                    match std::fs::write(PATH_FAN, &command) {
                        Ok(_) => {
                            self.current_level = level;
                            info!("Fan level set to {} after module reload", level);
                            Ok(())
                        }
                        Err(e2) => {
                            error!("Failed to set fan level after reload: {}", e2);
                            Err(anyhow::anyhow!("Failed to write to fan control after reload"))
                        }
                    }
                } else {
                    error!("Failed to set fan level: {}", e);
                    Err(anyhow::anyhow!("Failed to write to fan control"))
                }
            }
        }
    }

    fn set_auto_mode(&mut self) -> anyhow::Result<()> {
        let command = "level auto";
        match std::fs::write(PATH_FAN, command) {
            Ok(_) => {
                self.auto_mode = true;
                info!("Fan set to auto mode");
                Ok(())
            }
            Err(e) => {
                let error_str = e.to_string();
                if error_str.contains("Invalid argument") {
                    warn!("Invalid argument error, reloading thinkpad_acpi module");
                    self.reload_thinkpad_acpi()?;
                    
                    match std::fs::write(PATH_FAN, command) {
                        Ok(_) => {
                            self.auto_mode = true;
                            info!("Fan set to auto mode after module reload");
                            Ok(())
                        }
                        Err(e2) => {
                            error!("Failed to set auto mode after reload: {}", e2);
                            Err(anyhow::anyhow!("Failed to write to fan control after reload"))
                        }
                    }
                } else {
                    error!("Failed to set auto mode: {}", e);
                    Err(anyhow::anyhow!("Failed to write to fan control"))
                }
            }
        }
    }

    fn read_fan_status(&self) -> anyhow::Result<String> {
        let mut file = File::open(PATH_FAN)?;
        let mut buffer = [0; 64];
        let bytes_read = file.read(&mut buffer)?;
        
        let content = std::str::from_utf8(&buffer[..bytes_read])?;
        Ok(content.to_string())
    }

    fn reload_thinkpad_acpi(&self) -> anyhow::Result<()> {
        info!("Reloading thinkpad_acpi module");
        
        let unload_status = Command::new("modprobe")
            .args(["-r", "thinkpad_acpi"])
            .status();
            
        match unload_status {
            Ok(status) => {
                if !status.success() {
                    warn!("Failed to unload thinkpad_acpi module");
                }
            }
            Err(e) => {
                warn!("Error unloading thinkpad_acpi module: {}", e);
            }
        }
        
        std::thread::sleep(Duration::from_millis(1000));
        
        let load_status = Command::new("modprobe")
            .args(["thinkpad_acpi", "fan_control=1"])
            .status();
            
        match load_status {
            Ok(status) => {
                if status.success() {
                    info!("thinkpad_acpi module reloaded successfully");
                    std::thread::sleep(Duration::from_millis(2000));
                    Ok(())
                } else {
                    error!("Failed to reload thinkpad_acpi module");
                    Err(anyhow::anyhow!("Failed to reload thinkpad_acpi module"))
                }
            }
            Err(e) => {
                error!("Error reloading thinkpad_acpi module: {}", e);
                Err(anyhow::anyhow!("Error reloading thinkpad_acpi module"))
            }
        }
    }

    async fn run_loop(&mut self) -> anyhow::Result<()> {
        loop {
            if self.auto_mode {
                let adapters = self.read_temperatures();
                let max_temp = self.get_max_temp(&adapters);
                let new_level = self.calculate_fan_level(max_temp);
                
                if let Err(e) = self.set_fan_level(new_level) {
                    warn!("Failed to set fan level: {}", e);
                }
                
                info!("Max temp: {:.1}°C, Fan level: {}", max_temp, new_level);
            }
            
            sleep(Duration::from_secs(5)).await;
        }
    }
}

fn parse_adapters(json_str: &str) -> Vec<Adapter> {
    let json: serde_json::Value = match serde_json::from_str(json_str) {
        Ok(j) => j,
        Err(e) => {
            error!("Failed to parse JSON: {}", e);
            return vec![];
        }
    };

    let json_obj = match json.as_object() {
        Some(obj) => obj,
        None => {
            error!("Invalid JSON structure");
            return vec![];
        }
    };

    let mut adapters = Vec::new();

    for adapter_name in json_obj.keys() {
        let mut curr_inputs: Vec<(String, String)> = Vec::new();
        let adapter = match json_obj[adapter_name].as_object() {
            Some(adapter) => adapter,
            None => continue,
        };

        for input_key in adapter.keys() {
            let input = adapter[input_key].clone();
            if !input.is_object() {
                continue;
            }

            let input_obj = match input.as_object() {
                Some(obj) => obj,
                None => continue,
            };

            for temp_key in input_obj.keys() {
                if !temp_key.contains("temp") || !temp_key.contains("input") {
                    continue;
                }

                if input_obj[temp_key] == 0.0 {
                    continue;
                }

                curr_inputs.push((input_key.clone(), input_obj[temp_key].to_string()));
            }
        }

        if curr_inputs.is_empty() {
            continue;
        }

        let mut adapter = Adapter {
            name: adapter_name.clone(),
            inputs: Vec::new(),
        };

        for (name, temp) in curr_inputs {
            let temp_value = temp.trim().parse::<f64>().unwrap_or(-99.0);
            adapter.inputs.push(Input {
                name,
                temp: temp_value,
            });
        }
        adapters.push(adapter);
    }

    adapters
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    env_logger::init();
    
    let args = Args::parse();
    

    
    if !Path::new(PATH_FAN).exists() {
        error!("Fan control file not found: {}", PATH_FAN);
        return Err(anyhow::anyhow!("Fan control not available"));
    }

    let mut daemon = ThinkFanDaemon::new(&args.config)?;

    if args.auto {
        daemon.set_auto_mode()?;
        return Ok(());
    }

    if let Some(level) = args.level {
        if level > 7 {
            return Err(anyhow::anyhow!("Fan level must be 0-7"));
        }
        daemon.set_fan_level(level)?;
        return Ok(());
    }

    if args.daemon {
        info!("Starting thinkfand daemon");
        daemon.run_loop().await?;
    } else {
        let status = daemon.read_fan_status()?;
        println!("{}", status);
    }

    Ok(())
}
